""" fURI utilities """


import json
import os
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

import yaml
from . import furifile
from . import exceptions


def add_handler(scheme, cls):
    """ Add a new class to the dispatcher

        Arguments:
            scheme (str):    URI scheme prefix
            cls    (class):  Reference to class that extends furi.base """
    __dispatch__[scheme] = cls


def add_mapper(scheme, cls):
    """ Add a new class to the dispatcher

        Arguments:
            scheme (str):    URI scheme prefix
            cls    (class):  Reference to class that extends furi.base """
    __mapdispatch__[scheme] = cls


def add_mapext(ext, func):
    """ Add an extension handler for FileMap.

        Arguments:
            ext  (str):   File extension
            func (func):  Function to parse FileMap """
    __extdispatch__[ext] = func


def exists(uri, **kwargs):
    """ Returns True if URI exists. """
    with open(uri, **kwargs) as exister:
        return exister.exists()


def map(uri, **kwargs): # pylint: disable=redefined-builtin
    """ Return a Mapping object from a URI. """
    uri = urlparse(os.path.expanduser(uri))
    try:
        return __mapdispatch__[uri.scheme](uri.geturl(), **kwargs)
    except KeyError:
        raise exceptions.SchemeError("Unsupported URI scheme: '%s'" % uri.scheme)


def open(uri, **kwargs): # pylint: disable=redefined-builtin
    """ Returns a File object given a URI. """
    uri = urlparse(os.path.expanduser(uri))
    try:
        return __dispatch__[uri.scheme](uri.geturl(), **kwargs)
    except KeyError:
        raise exceptions.SchemeError("Unsupported URI scheme: '%s'" % uri.scheme)


def walk(uri, **kwargs):
    """ Walk a Directory given a URI. """
    with open(uri, **kwargs) as walker:
        return walker.walk()


def download(source, target=None, **credentials):
    """ Download contents of a source URI into a target URI. If target URI is omitted,
        source is downloaded to ~/Downloads using the same filename.

        Arguments:
            source      (str):   URI of source file
            target      (str):   URI of target file (optional)
            credentials (dict):  Optional source connection credentials

        Returns:
            Handle to target file """
    src = open(source)
    src.connect(**credentials)
    tgt = open(target or os.path.expanduser("~/Downloads/%s" % src.filename), mode='r+')

    if not isinstance(src, furifile.RemoteFile):
        raise exceptions.DownloadError("Cannot download from non-RemoteFile.")
    if isinstance(tgt, furifile.RemoteFile):
        raise exceptions.DownloadError(
            "Cannot download RemoteFile to other RemoteFile. Use local URI.")

    return src.download(tgt)


__dispatch__ = {
    '': furifile.File,
    'file': furifile.File
}


__mapdispatch__ = {}


__extdispatch__ = {
    '.json' : json.loads,
    '.yaml' : yaml.load,
    '.yml'  : yaml.load
}



import io
import os
import re
import tempfile
import boto
import furi
import moto
from nose.tools import assert_equal
from nose.tools import assert_is_instance


def test_matches():
    pattern = re.compile('[A-Z][a-z]+\.[Yy][Aa]?[Mm][Ll]')
    furifile = furi.open('/abs/path/to/Hello.yml')
    assert furifile.matches(pattern)


def test_not_matches():
    pattern = re.compile('[A-Z][a-z]+\.[Yy][Aa]?[Mm][Ll]')
    furifile = furi.open('/abs/path/to/Hello.txt')
    assert not furifile.matches(pattern)


def test_exists():
    furifile = furi.open(__file__)
    assert furifile.exists()


def test_not_exists():
    furifile = furi.open('/foo/bar/fizz/buzz')
    assert not furifile.exists()


def test_read():
    with open(__file__, 'r') as this:
        furifile = furi.open(__file__)
        returned = furifile.read()
        expected = this.read()
        assert_equal(returned, expected)


def test_stream_resets():
    with open(__file__, 'r') as this:
        furifile = furi.open(__file__)
        stream = furifile.stream()
        stream.read()
        returned = furifile.read()
        expected = this.read()
        assert_equal(returned, expected)


def test_write():
    value = "Hello, world!\n\nGoodby, cruel world."
    with tempfile.NamedTemporaryFile() as tmp:
        furifile = furi.open(tmp.name, mode='w+')
        furifile.write(value)
        furifile.stream().seek(0)
        assert_equal(furifile.read(), value)


@moto.mock_s3
def test_download():
    value = "Hello, world!\n\nGoodby, cruel world."
    ms3   = boto.connect_s3()
    bkt   = ms3.create_bucket('furi')
    key   = boto.s3.key.Key(bkt, 'foo/bar/bizz/buzz')
    key.set_contents_from_string(value)

    with tempfile.NamedTemporaryFile() as tmp:
        furifile = furi.download('s3://furi/foo/bar/bizz/buzz', tmp.name)
        assert_equal(furifile.read(), value)


@moto.mock_s3
def test_s3_stream():
    value = "Hello, world!\n\nGoodby, cruel world."
    ms3   = boto.connect_s3()
    bkt   = ms3.create_bucket('furi')
    key   = boto.s3.key.Key(bkt, 'foo/bar/bizz/buzz')
    key.set_contents_from_string(value)

    with furi.open('s3://furi/foo/bar/bizz/buzz') as tmp:
        tmp.stream()
        tmp.stream()


@moto.mock_s3
def test_exists():
    value = "Hello, world!\n\nGoodby, cruel world."
    ms3   = boto.connect_s3()
    bkt   = ms3.create_bucket('furi')
    key   = boto.s3.key.Key(bkt, 'foo/bar/bizz/buzz')
    key.set_contents_from_string(value)

    s3furi = furi.open('s3://furi/foo/bar/bizz/buzz')
    assert s3furi.exists()


@moto.mock_s3
def test_not_exists():
    ms3 = boto.connect_s3()
    bkt = ms3.create_bucket('furi')

    s3furi = furi.open('s3://furi/foo/bar/bizz/buzz')
    assert not s3furi.exists()


@moto.mock_s3
def test_write():
    ms3 = boto.connect_s3()
    bkt = ms3.create_bucket('furi')
    key = boto.s3.key.Key(bkt)
    key.name = 'foo/bar/bizz/buzz'
    value    = "Hello, world!\n\nGoodby, cruel world."
    key.set_contents_from_string(value)

    with furi.open('s3://furi/foo/bar/bizz/buzz', mode='w') as s3furi:
        yield assert_equal, s3furi.read(), value


@moto.mock_s3
def test_write_stream():
    ms3 = boto.connect_s3()
    bkt = ms3.create_bucket('furi')
    key = boto.s3.key.Key(bkt)
    key.name = 'foo/bar/bizz/buzz'
    value    = "Hello, world!\n\nGoodby, cruel world."
    key.set_contents_from_stream(io.StringIO(unicode(value)))

    with furi.open('s3://furi/foo/bar/bizz/buzz', mode='w') as s3furi:
        yield assert_equal, s3furi.read(), value

@moto.mock_s3
def test_s3_walk():
    value = "Hello, world!\n\nGoodby, cruel world."
    ms3   = boto.connect_s3()
    bkt   = ms3.create_bucket('furi')

    for keyname in ['foo/baq/bug', 'foo/bar/bizz/buzz', 'foo/bar/bizz/fizz', 'foo/ban']:
        key = boto.s3.key.Key(bkt)
        key.name = keyname
        key.set_contents_from_string(value)

    returned = list(furi.walk('s3://furi/foo/'))
    expected = [
        ('s3://furi/foo/',          ['baq', 'bar'],  ['ban']),
        ('s3://furi/foo/baq/',      [],              ['bug']),
        ('s3://furi/foo/bar/',      ['bizz'],        []),
        ('s3://furi/foo/bar/bizz/', [],              ['buzz', 'fizz']) ]
    assert_equal(returned, expected)

def test_walk():
    with tempfile.NamedTemporaryFile() as tmp:
        path = os.path.split(tmp.name)[0]
        os.chdir(path)
        for keyname in ['foo/baq/bug', 'foo/bar/bizz/buzz', 'foo/bar/bizz/fizz', 'foo/ban']:
            fpath, fname = os.path.split(keyname)
            if not os.path.exists(fpath):
                os.makedirs(fpath)
            with open(keyname, 'w') as tmpf:
                tmpf.write("Hello, world!\n\nGoodby, cruel world.")

        returned = list(furi.walk('foo'))
        expected = [
            ('foo',          ['baq', 'bar'],  ['ban']),
            ('foo/baq',      [],              ['bug']),
            ('foo/bar',      ['bizz'],        []),
            ('foo/bar/bizz', [],              ['buzz', 'fizz']) ]
        assert_equal(returned, expected)



""" fURI File access through URIs. """

from . import exceptions
from .furimap import chain
from .utils import add_handler
from .utils import add_mapper
from .utils import add_mapext
from .utils import download
from .utils import exists
from .utils import map  # pylint: disable=redefined-builtin
from .utils import open # pylint: disable=redefined-builtin
from .utils import walk
try:
    from . import aws
except ImportError:
    pass
try:
    from . import sftp
except ImportError:
    pass


__author__ = "amancevice"
__email__ = "smallweirdnum@gmail.com"
__version__ = "0.7.0"



""" SFTP backed File implementation. """

import io
import os
import re
import tempfile

import pysftp
from . import furifile
from . import utils


class SftpFile(furifile.RemoteFile):
    """ File on SFTP server.

        Ex. sftp://user:pass@host/path/to/file.ext """

    def _connect(self):
        """ Connect to remote. Credentials can be passed as part of URI. """
        # Get user, pswd, and host from URI
        ptn = r'^((?P<user>.*?)(:(?P<pswd>.*))?@)?(?P<host>.*?):?(?P<port>\d+)?$'
        match = re.match(ptn, self.uri.netloc)
        user = match.group('user')
        pswd = match.group('pswd')
        host = match.group('host')
        port = int(match.group('port') or '22')

        self.__connect__.setdefault('username', user)
        self.__connect__.setdefault('password', pswd)
        self.__connect__.setdefault('port', port)

        return pysftp.Connection(host, **self.__connect__)

    def _download(self, target):
        """ Download remote file to local target URI. """
        self.connection.get(self.filename, str(target))
        return target

    def _exists(self):
        """ Test file existence. """
        return self.connection.exists(self.path)

    def _write(self, stream):
        """ Write stream to file. """
        raise NotImplementedError

    def _stream(self):
        """ Implementation of stream(). """
        with tempfile.NamedTemporaryFile() as tmp:
            self.connection.get(self.path, tmp.name)
            tmp.flush()
            try:
                return io.StringIO(unicode(tmp.read()))
            except NameError:
                io.StringIO(tmp.read())


    def _walk(self):
        """ Implementation of walk(). """
        def __walk(dirpath):
            """ Iterator for _walk_impl(). """
            wtcb = pysftp.WTCallbacks()
            self.connection.walktree(
                dirpath, wtcb.file_cb, wtcb.dir_cb, wtcb.unk_cb, recurse=False)
            yield dirpath, wtcb.dlist, [os.path.split(x)[-1] for x in wtcb.flist]
            for dirname in wtcb.dlist:
                for dirpath, dirnames, filenames in __walk(dirname):
                    yield dirpath, dirnames, filenames
        return __walk(".%s" % self.path)


utils.add_handler('sftp', SftpFile)



""" fURI Exceptions """


class DownloadError(TypeError):
    """ Error downloading object. """
    pass


class ExtensionError(KeyError):
    """ Error mapping file from extension. """
    pass


class FuriFileNotFoundError(ValueError):
    """ File does not exist. """
    pass


class ModeError(ValueError):
    """ File cannot be opened in given mode. """
    pass


class SchemeError(KeyError):
    """ Error finding class to open fURI. """
    pass



""" fURI Files. """

import collections
import os
import re
try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

from . import exceptions


class File(collections.Iterable):
    """ Local File implementation.

        Exs. file:///abs/path/to/file.ext
             /abs/path/to/file.ext """

    __modes__ = {'r', 'rb', 'r+', 'rb+', 'w', 'wb', 'w+', 'wb+', 'a', 'ab', 'a+', 'ab+'}

    def __init__(self, uri, mode='r'):
        if mode not in self.__modes__ and not set(mode).issubset(self.__modes__):
            raise exceptions.ModeError(
                "Cannot open %s in %s-mode" % (type(self).__name__, mode))
        self.uri = urlparse(uri)
        self.path = self.uri.path
        self.mode = mode
        self.workdir, self.filename = os.path.split(self.path)
        self.__stream__ = None

    def __str__(self):
        return self.uri.geturl()

    def __repr__(self):
        return "<%s: %s>" % (type(self).__name__, str(self))

    def __iter__(self):
        return iter(self.stream())

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """ Close stream. """
        return self._close()

    def exists(self):
        """ Test file existence. """
        return self._exists()

    def matches(self, pattern):
        """ Filename matches pattern.

            Arguments:
                pattern (regex or str):  Regular Expression to match.

            Returns:
                RegEx match object. """
        return re.compile(pattern).match(self.filename)

    def read(self, *size):
        """ Read file stream. """
        return self._read(*size)

    def stream(self):
        """ Get file contents as stream. """
        if self.__stream__ is not None and hasattr(self.__stream__, 'seek'):
            self.__stream__.seek(0)
        else:
            if not self.exists() and 'w' not in self.mode:
                raise exceptions.FuriFileNotFoundError("%s does not exist" % self.uri.geturl())
            self.__stream__ = self._stream()
        return self.__stream__

    def walk(self, **kwargs):
        """ Walk the contents of a directory. """
        return self._walk(**kwargs)

    def write(self, stream):
        """ Write stream. """
        self._write(stream)

    def _close(self):
        """ Close stream implementation. """
        try:
            self.__stream__.close()
        except:  # pylint: disable=bare-except
            pass
        finally:
            self.__stream__ = None

    def _exists(self):
        """ Test file existence implementation. """
        return os.path.exists(self.path)

    def _read(self, *size):
        """ Read file stream implementation. """
        return self.stream().read(*size)

    def _stream(self):
        """ Implementation of stream(). """
        return open(self.path, self.mode)

    def _walk(self, **kwargs):
        """ Implementation of walk(). """
        return os.walk(self.path, **kwargs)

    def _write(self, stream):
        """ Write stream implementation. """
        if not os.path.exists(self.workdir) and ('w' in self.mode or 'a' in self.mode):
            os.makedirs(self.workdir)
        try:
            return self.stream().write(stream.read())
        except AttributeError:
            return self.stream().write(stream)


class RemoteFile(File):
    """ Remote file implementation. """
    def __init__(self, uri, mode='r', **connectkw):
        super(RemoteFile, self).__init__(uri, mode=mode)
        self.__connect__ = connectkw
        self.__connection__ = None

    @property
    def connection(self):
        """ Remote connection. """
        if self.__connection__ is None:
            self.connect()
        return self.__connection__

    def connect(self, **connectkw):
        """ Connect to remote. """
        if any(connectkw):
            self.__connect__ = connectkw
        self.__connection__ = self._connect()

    def download(self, target):
        """ Download remote file to local target URI. """
        return self._download(target)

    def _connect(self):
        """ Connect to remote implementation. """
        raise NotImplementedError

    def _download(self, target):
        """ Download remote file to local target URI implementation. """
        raise NotImplementedError

    def _exists(self):
        """ Test file existence. """
        raise NotImplementedError

    def _write(self, stream):
        """ Write stream to file. """
        raise NotImplementedError

    def _walk(self, **kwargs):
        """ Implementation of walk(). """
        raise NotImplementedError

    def _stream(self):
        """ Implementation of stream(). """
        raise NotImplementedError
